@@ -191,13 +191,13 @@ gnutls_compression_set_priority(gnutls_session_t session, const int *list)
  **/
 int gnutls_protocol_set_priority(gnutls_session_t session, const int *list)
 {
-	_set_priority(&session->internals.priorities.protocol, list);
-
-	/* set the current version to the first in the chain.
-	 * This will be overridden later.
-	 */
-	if (list)
+	if (list) {
+		_set_priority(&session->internals.priorities.protocol, list);
+		/* set the current version to the first in the chain.
+		 * This will be overridden later.
+		 */
 		_gnutls_set_current_version(session, list[0]);
+	}
 
 	return 0;
 }
@@ -602,6 +602,10 @@ gnutls_priority_set(gnutls_session_t session, gnutls_priority_t priority)
 #define LEVEL_SUITEB192 "SUITEB192"
 #define LEVEL_EXPORT "EXPORT"
 
+#define SET_LEVEL(to_set) \
+	if (priority_cache->level == 0 || priority_cache->level > to_set) \
+		priority_cache->level = to_set
+
 static
 int check_level(const char *level, gnutls_priority_t priority_cache,
 		int add)
@@ -620,8 +624,7 @@ int check_level(const char *level, gnutls_priority_t priority_cache,
 		func(&priority_cache->sign_algo, sign_priority_default);
 		func(&priority_cache->supported_ecc, supported_ecc_normal);
 
-		if (priority_cache->level == 0)
-			priority_cache->level = GNUTLS_SEC_PARAM_VERY_WEAK;
+		SET_LEVEL(GNUTLS_SEC_PARAM_VERY_WEAK);
 		return 1;
 	} else if (strcasecmp(level, LEVEL_NORMAL) == 0) {
 		func(&priority_cache->cipher, cipher_priority_normal);
@@ -630,8 +633,7 @@ int check_level(const char *level, gnutls_priority_t priority_cache,
 		func(&priority_cache->sign_algo, sign_priority_default);
 		func(&priority_cache->supported_ecc, supported_ecc_normal);
 
-		if (priority_cache->level == 0)
-			priority_cache->level = GNUTLS_SEC_PARAM_VERY_WEAK;
+		SET_LEVEL(GNUTLS_SEC_PARAM_VERY_WEAK);
 		return 1;
 	} else if (strcasecmp(level, LEVEL_PFS) == 0) {
 		func(&priority_cache->cipher, cipher_priority_normal);
@@ -640,8 +642,7 @@ int check_level(const char *level, gnutls_priority_t priority_cache,
 		func(&priority_cache->sign_algo, sign_priority_default);
 		func(&priority_cache->supported_ecc, supported_ecc_normal);
 
-		if (priority_cache->level == 0)
-			priority_cache->level = GNUTLS_SEC_PARAM_VERY_WEAK;
+		SET_LEVEL(GNUTLS_SEC_PARAM_VERY_WEAK);
 		return 1;
 	} else if (strcasecmp(level, LEVEL_SECURE256) == 0
 		   || strcasecmp(level, LEVEL_SECURE192) == 0) {
@@ -653,8 +654,7 @@ int check_level(const char *level, gnutls_priority_t priority_cache,
 		     supported_ecc_secure192);
 
 		/* be conservative for now. Set the bits to correspond to 96-bit level */
-		if (priority_cache->level == 0)
-			priority_cache->level = GNUTLS_SEC_PARAM_LEGACY;
+		SET_LEVEL(GNUTLS_SEC_PARAM_LEGACY);
 		return 1;
 	} else if (strcasecmp(level, LEVEL_SECURE128) == 0
 		   || strcasecmp(level, "SECURE") == 0) {
@@ -666,8 +666,7 @@ int check_level(const char *level, gnutls_priority_t priority_cache,
 		     supported_ecc_secure128);
 
 		/* be conservative for now. Set the bits to correspond to an 72-bit level */
-		if (priority_cache->level == 0)
-			priority_cache->level = GNUTLS_SEC_PARAM_WEAK;
+		SET_LEVEL(GNUTLS_SEC_PARAM_WEAK);
 		return 1;
 	} else if (strcasecmp(level, LEVEL_SUITEB128) == 0) {
 		func(&priority_cache->protocol, protocol_priority_suiteb);
@@ -678,8 +677,7 @@ int check_level(const char *level, gnutls_priority_t priority_cache,
 		func(&priority_cache->supported_ecc,
 		     supported_ecc_suiteb128);
 
-		if (priority_cache->level == 0)
-			priority_cache->level = GNUTLS_SEC_PARAM_HIGH;
+		SET_LEVEL(GNUTLS_SEC_PARAM_HIGH);
 		return 1;
 	} else if (strcasecmp(level, LEVEL_SUITEB192) == 0) {
 		func(&priority_cache->protocol, protocol_priority_suiteb);
@@ -690,8 +688,7 @@ int check_level(const char *level, gnutls_priority_t priority_cache,
 		func(&priority_cache->supported_ecc,
 		     supported_ecc_suiteb192);
 
-		if (priority_cache->level == 0)
-			priority_cache->level = GNUTLS_SEC_PARAM_ULTRA;
+		SET_LEVEL(GNUTLS_SEC_PARAM_ULTRA);
 		return 1;
 	} else if (strcasecmp(level, LEVEL_EXPORT) == 0) {
 		func(&priority_cache->cipher, cipher_priority_performance);
@@ -700,8 +697,7 @@ int check_level(const char *level, gnutls_priority_t priority_cache,
 		func(&priority_cache->sign_algo, sign_priority_default);
 		func(&priority_cache->supported_ecc, supported_ecc_normal);
 
-		if (priority_cache->level == 0)
-			priority_cache->level = GNUTLS_SEC_PARAM_EXPORT;
+		SET_LEVEL(GNUTLS_SEC_PARAM_EXPORT);
 		return 1;
 	}
 	return 0;
@@ -788,6 +784,7 @@ gnutls_priority_init(gnutls_priority_t * priority_cache,
 	char *broken_list[MAX_ELEMENTS];
 	int broken_list_size = 0, i = 0, j;
 	char *darg = NULL;
+	unsigned ikeyword_set = 0;
 	int algo;
 	rmadd_func *fn;
 	bulk_rmadd_func *bulk_fn;
@@ -838,11 +835,13 @@ gnutls_priority_init(gnutls_priority_t * priority_cache,
 			      supported_ecc_normal);
 		i = 0;
 	} else {
+		ikeyword_set = 1;
 		i = 1;
 	}
 
 	for (; i < broken_list_size; i++) {
-		if (check_level(broken_list[i], *priority_cache, 0) != 0) {
+		if (check_level(broken_list[i], *priority_cache, ikeyword_set) != 0) {
+			ikeyword_set = 1;
 			continue;
 		} else if (broken_list[i][0] == '!'
 			   || broken_list[i][0] == '+'
@@ -1079,6 +1078,7 @@ gnutls_priority_init(gnutls_priority_t * priority_cache,
 	}
 	gnutls_free(darg);
 	gnutls_free(*priority_cache);
+	*priority_cache = NULL;
 
 	return GNUTLS_E_INVALID_REQUEST;
 
